"""

Submodules
==========

.. autosummary::
    :toctree: _autosummary


"""

__all__ = (
    "CarInputParameters",
    "fill_xarray_from_input_parameters",
    "CarModel",
    "InventoryCar",
    "get_standard_driving_cycle_and_gradient",
)
__version__ = (1, 9, 0)

from pathlib import Path

DATA_DIR = Path(__file__).resolve().parent / "data"

from carculator_utils.array import fill_xarray_from_input_parameters
from carculator_utils.driving_cycles import get_standard_driving_cycle_and_gradient

from .car_input_parameters import CarInputParameters
from .inventory import InventoryCar
from .model import CarModel



import numpy as np
from carculator_utils.energy_consumption import EnergyConsumptionModel


def test_acceleration():
    ecm = EnergyConsumptionModel(
        vehicle_type="car",
        vehicle_size=["Medium"],
        cycle="WLTC",
        gradient="WLTC",
        powertrains=["ICEV-d"],
    )
    assert ecm.cycle.shape == ecm.gradient.shape


def test_aux_power():
    ecm = EnergyConsumptionModel(
        vehicle_type="car",
        vehicle_size=["Medium"],
        cycle="WLTC",
        gradient="WLTC",
        powertrains=["ICEV-d"],
    )
    aux = ecm.aux_energy_per_km(aux_power=np.array([100]), efficiency=np.array([0.5]))
    assert isinstance(aux, np.ndarray)
    assert np.allclose(aux, np.repeat(100 / 0.5, 3144))



import json
from pathlib import Path

import carculator.car_input_parameters as vip

DEFAULT = Path(__file__, "..").resolve() / "fixtures" / "default_test.json"
EXTRA = Path(__file__, "..").resolve() / "fixtures" / "extra_test.json"


def test_can_pass_directly():
    d, e = json.load(open(DEFAULT)), set(json.load(open(EXTRA)))
    e.remove("foobazzle")
    assert len(vip.CarInputParameters(d, e).powertrains) == 12
    assert len(vip.CarInputParameters(d, e).parameters) == 332


def test_alternate_filepath():
    assert len(vip.CarInputParameters(DEFAULT, EXTRA).powertrains) == 12
    assert len(vip.CarInputParameters(DEFAULT, EXTRA).parameters) == 332



from pathlib import Path

import numpy as np
import pandas as pd

from carculator import *

DATA = Path(__file__, "..").resolve() / "fixtures" / "cars_values.xlsx"
OUTPUT = Path(__file__, "..").resolve() / "fixtures" / "test_model_results.xlsx"
ref = pd.read_excel(DATA, index_col=0)

cip = CarInputParameters()
cip.static()
dcts, arr = fill_xarray_from_input_parameters(cip)
cm = CarModel(arr, cycle="WLTC")
cm.set_all()


def test_model_results():
    list_powertrains = [
        "ICEV-p",
        "ICEV-d",
        "PHEV-p",
        "PHEV-d",
        "BEV",
        "ICEV-g",
        "HEV-d",
        "HEV-p",
    ]
    list_sizes = [
        # "Small",
        # "Lower medium",
        "Medium",
        # "Large"
    ]
    list_years = [
        2020,
        # 2030,
        # 2040,
        # 2050
    ]

    l_res = []

    for pwt in list_powertrains:
        for size in list_sizes:
            for year in list_years:
                for param in cm.array.parameter.values:
                    val = float(
                        cm.array.sel(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter=param,
                            value=0,
                        ).values
                    )

                    try:
                        ref_val = (
                            ref.loc[
                                (ref["powertrain"] == pwt)
                                & (ref["size"] == size)
                                & (ref["parameter"] == param),
                                year,
                            ]
                            .values.astype(float)
                            .item(0)
                        )
                    except:
                        ref_val = 1

                    _ = lambda x: np.where(ref_val == 0, 1, ref_val)
                    diff = val / _(ref_val)
                    l_res.append([pwt, size, year, param, val, ref_val, diff])

    pd.DataFrame(
        l_res,
        columns=["powertrain", "size", "year", "parameter", "val", "ref_val", "diff"],
    ).to_excel(OUTPUT)


def test_setting_batt_cap():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV"], "year": [2020]},
    )

    batt_cap = {
        "capacity": {
            ("BEV", "Medium", 2020): 50,
        }
    }

    cm = CarModel(arr, cycle="WLTC", energy_storage=batt_cap)
    cm.set_all()

    assert (
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="electric energy stored",
            value=0,
        ).values
        == 50
    )


def test_setting_battery_chemistry():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV"], "year": [2020]},
    )

    batt_chem = {
        "electric": {
            ("BEV", "Medium", 2020): "LFP",
        }
    }

    cm = CarModel(arr, cycle="WLTC", energy_storage=batt_chem)
    cm.set_all()

    assert cm.array.sel(
        powertrain="BEV",
        size="Medium",
        year=2020,
        parameter="battery cell energy density",
        value=0,
    ).values == np.array(0.15, dtype=np.float32)


def test_setting_range():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV"], "year": [2020]},
    )

    range = {
        ("BEV", "Medium", 2020): 100,
    }

    cm = CarModel(arr, cycle="WLTC", target_range=range)
    cm.set_all()

    assert (
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="electric energy stored",
            value=0,
        ).values
        <= 25
    )

    assert np.isclose(
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="range",
            value=0,
        ).values,
        np.array(100, dtype=np.float32),
        rtol=0.01,
    )


def test_setting_mass():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV"], "year": [2020]},
    )

    mass = {
        ("BEV", "Medium", 2020): 2000,
    }

    cm = CarModel(arr, cycle="WLTC", target_mass=mass)
    cm.set_all()

    assert (
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="curb mass",
            value=0,
        ).values
        == 2000
    )


def test_setting_ttw_energy():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV", "ICEV-p"], "year": [2020]},
    )

    ttw_energy = {
        ("BEV", "Medium", 2020): 1000,
        ("ICEV-p", "Medium", 2020): 2500,
    }

    cm = CarModel(arr, cycle="WLTC", energy_consumption=ttw_energy)
    cm.set_all()

    assert (
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="TtW energy",
            value=0,
        ).values
        == 1000
    )

    assert np.isclose(
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="electricity consumption",
            value=0,
        ).values,
        (1000 / 3600) * 1.17,
        rtol=0.01,
    )

    assert (
        cm.array.sel(
            powertrain="ICEV-p",
            size="Medium",
            year=2020,
            parameter="TtW energy",
            value=0,
        ).values
        == 2500
    )

    _ = lambda x: np.where(x == 0, 1, x)
    assert np.array_equal(
        cm["fuel consumption"],
        (cm["fuel mass"] / (cm["range"]) / _(cm["fuel density per kg"])),
    )


def test_setting_power():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV", "ICEV-p"], "year": [2020]},
    )

    power = {
        ("BEV", "Medium", 2020): 100,
        ("ICEV-p", "Medium", 2020): 200,
    }

    cm = CarModel(arr, cycle="WLTC", power=power)
    cm.set_all()

    assert (
        cm.array.sel(
            powertrain="BEV",
            size="Medium",
            year=2020,
            parameter="electric power",
            value=0,
        ).values
        == 100
    )

    assert (
        cm.array.sel(
            powertrain="ICEV-p",
            size="Medium",
            year=2020,
            parameter="combustion power",
            value=0,
        ).values
        == 200
    )

    assert (
        cm.array.sel(
            powertrain="ICEV-p",
            size="Medium",
            year=2020,
            parameter="power",
            value=0,
        ).values
        == 200
    )

    assert np.array_equal(
        cm["combustion engine mass"],
        (
            cm["combustion power"] * cm["combustion mass per power"]
            + cm["combustion fixed mass"]
        ),
    )


def test_setting_battery_origin():
    cip = CarInputParameters()
    cip.static()
    dcts, arr = fill_xarray_from_input_parameters(
        cip,
        scope={"size": ["Medium"], "powertrain": ["BEV"], "year": [2020]},
    )

    battery_origin = {"origin": "FR"}

    cm = CarModel(arr, cycle="WLTC", energy_storage=battery_origin)
    cm.set_all()

    assert cm.energy_storage["origin"] == "FR"



"""
inventory.py contains Inventory which provides all methods to solve inventories.
"""

import warnings

import numpy as np
from carculator_utils.inventory import Inventory, format_array

from . import DATA_DIR

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

IAM_FILES_DIR = DATA_DIR / "IAM"


class InventoryCar(Inventory):
    """
    Build and solve the inventory for results
    characterization and inventory export

    """

    def fill_in_A_matrix(self):
        """
        Fill-in the A matrix. Does not return anything. Modifies in place.
        Shape of the A matrix (values, products, activities).

        :param array: :attr:`array` from :class:`CarModel` class
        """

        # Glider
        self.A[
            :,
            self.find_input_indices(("market for glider, passenger car",)),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="glider base mass") * -1
        )

        self.A[
            :,
            self.find_input_indices(("glider lightweighting",)),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="lightweighting")
            * self.array.sel(parameter="glider base mass")
            * -1
        )

        self.A[
            :,
            self.find_input_indices(("maintenance, passenger car",)),
            self.find_input_indices(("transport, car, ",)),
        ] = (
            self.array.sel(parameter="curb mass") / 1240 / 150000 * -1
        )

        # Fuel tank EoL
        self.A[
            :,
            self.find_input_indices(
                ("market for waste plastic, industrial electronics",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = self.array.sel(parameter="fuel tank mass")

        # EoL Glider
        # the EoL treatment of the glider is already
        # taken into account in the glider production dataset
        # we do not consider the reduced EoL due to lightweighting

        # Powertrain components
        self.A[
            :,
            self.find_input_indices(("market for charger, electric passenger car",)),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="charger mass") * -1
        )

        self.A[
            :,
            self.find_input_indices(
                ("market for converter, for electric passenger car",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="converter mass") * -1
        )

        self.A[
            :,
            self.find_input_indices(
                ("market for electric motor, electric passenger car",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="electric engine mass") * -1
        )

        self.A[
            :,
            self.find_input_indices(
                ("market for inverter, for electric passenger car",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="inverter mass") * -1
        )

        self.A[
            :,
            self.find_input_indices(
                ("market for power distribution unit, for electric passenger car",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter="power distribution unit mass") * -1
        )

        l_elec_pt = [
            "charger mass",
            "converter mass",
            "inverter mass",
            "power distribution unit mass",
            # "powertrain mass",
            "electric engine mass",
            "fuel cell stack mass",
            "fuel cell ancillary BoP mass",
            "fuel cell essential BoP mass",
        ]

        self.A[
            :,
            self.find_input_indices(
                (
                    "market for used powertrain from electric passenger car, manual dismantling",
                )
            ),
            self.find_input_indices(("Car, ",)),
        ] = self.array.sel(parameter=l_elec_pt).sum(dim="parameter")

        self.A[
            :,
            self.find_input_indices(
                ("market for internal combustion engine, passenger car",)
            ),
            self.find_input_indices(("Car, ",)),
        ] = (
            self.array.sel(parameter=["combustion engine mass", "powertrain mass"]).sum(
                dim="parameter"
            )
            * -1
        )

        # EoL internal combustion engine
        # is already taken into account
        # in the ICE production dataset

        # Energy storage
        self.add_fuel_cell_stack()
        self.add_hydrogen_tank()
        self.add_battery()

        self.A[
            :,
            self.find_input_indices(
                contains=("polyethylene production, high density, granulate",)
            ),
            self.find_input_indices(
                contains=("Car, ",),
            ),
        ] = (
            self.array.sel(parameter="fuel tank mass")
            * (self.array.sel(parameter="combustion power") > 0)
            * (self.array.sel(parameter="CNG tank mass intercept") == 0)
            * -1
        )

        self.add_cng_tank()

        # END of vehicle building

        # Add vehicle dataset to transport dataset
        self.add_vehicle_to_transport_dataset()

        self.display_renewable_rate_in_mix()

        self.add_electricity_to_electric_vehicles()

        self.add_hydrogen_to_fuel_cell_vehicles()

        self.add_fuel_to_vehicles("methane", ["ICEV-g"], "EV-g")

        # Gas leakage to air
        self.A[
            :,
            self.find_input_indices(("fuel supply for methane vehicles",)),
            self.find_input_indices((f"transport, {self.vm.vehicle_type}, ",)),
        ] *= 1 + self.array.sel(parameter="CNG pump-to-tank leakage")

        # Gas leakage to air
        self.A[
            :,
            self.inputs[("Methane, fossil", ("air",), "kilogram")],
            self.find_input_indices((f"transport, {self.vm.vehicle_type}, ",)),
        ] *= self.array.sel(parameter="CNG pump-to-tank leakage")

        self.add_fuel_to_vehicles("diesel", ["ICEV-d", "PHEV-d", "HEV-d"], "EV-d")

        self.add_fuel_to_vehicles("petrol", ["ICEV-p", "PHEV-p", "HEV-p"], "EV-p")

        self.add_abrasion_emissions()

        self.add_road_construction()

        self.add_road_maintenance()

        self.add_exhaust_emissions()

        self.add_noise_emissions()

        self.add_refrigerant_emissions()

        print("*********************************************************************")



"""
car_input_parameters.py contains the VehicleInputParameters class, which, after instantiated, contains
all definitions (metadata) of input and calculated parameters, along with their values.
Also, it inherits methods from `klausen`, which exposes the methods .static() and .stochastic(),
which generate single or random values for input parameters.
"""

import json
from pathlib import Path
from typing import Union

from carculator_utils.vehicle_input_parameters import VehicleInputParameters


def load_parameters(obj: Union[str, Path, list]) -> list:
    """
    Returns a json object containing parameters' definitions
    :param obj: A filepath to a json file, or a json object
    :return: Returns a json object containing parameters' definitions
    """
    if isinstance(obj, (str, Path)):
        assert Path(obj).exists(), f"Can't find this filepath {obj}."
        return json.load(open(obj, encoding="utf-8"))

    # Already in correct form, just return
    return obj


class CarInputParameters(VehicleInputParameters):
    """ """

    DEFAULT = Path(__file__, "..").resolve() / "data" / "default_parameters.json"
    EXTRA = Path(__file__, "..").resolve() / "data" / "extra_parameters.json"

    def __init__(
        self,
        parameters: Union[str, Path, list] = None,
        extra: Union[str, Path, list] = None,
    ) -> None:
        """Create a `klausen <https://github.com/cmutel/klausen>`__ model with the car input parameters."""
        super().__init__(None)



from itertools import product

import numpy as np
import yaml
from carculator_utils.energy_consumption import EnergyConsumptionModel
from carculator_utils.model import VehicleModel

from . import DATA_DIR


class CarModel(VehicleModel):
    """
    This class represents the entirety of the vehicles considered, with useful attributes, such as an array that stores
    all the vehicles parameters.

    :ivar array: multi-dimensional numpy-like array that contains parameters' value(s)
    :ivar cycle: name of a driving cycle, or custom driving cycle
    :ivar gradient: series of gradients, for each second of the driving cycle
    :ivar energy_storage: dictionary with selection of battery chemistry for each powertrain

    """

    def set_all(self):
        """
        This method runs a series of other methods to obtain the tank-to-wheel energy requirement, efficiency
        of the car, costs, etc.

        :meth:`set_component_masses()`, :meth:`set_car_masses()` and :meth:`set_power_parameters()` are interdependent.
        `powertrain_mass` depends on `power`, `curb_mass` is affected by changes in `powertrain_mass`,
        `combustion engine mass` and `electric engine mass`, and `power` is a function of `curb_mass`.
        The current solution is to loop through the methods until the increment in driving mass is
        inferior to 0.1%.

        :param drop_hybrids: boolean. True by default. If False, the underlying vehicles used to build plugin-hybrid
                vehicles remain present in the array.
        :param electric_utility_factor: array. If an array is passed, its values are used to override the
                electric utility factor for plugin hybrid vehicles. If not, this factor is calculated using a relation
                described in `set_electric_utility_factor()`

        :returns: Does not return anything. Modifies ``self.array`` in place.

        """

        self.ecm = EnergyConsumptionModel(
            vehicle_type="car",
            vehicle_size=list(self.array.coords["size"].values),
            powertrains=list(self.array.coords["powertrain"].values),
            cycle=self.cycle,
            gradient=self.gradient,
            country=self.country,
        )

        diff = 1.0

        while diff > 0.00001:
            old_driving_mass = self["driving mass"].sum().values
            self.set_vehicle_mass()
            self.set_power_parameters()
            self.set_component_masses()
            self.set_auxiliaries()
            self.set_power_battery_properties()
            self.set_battery_properties()
            self.set_energy_stored_properties()
            self.set_recuperation()

            if "FCEV" in self.array.powertrain.values:
                self.set_fuel_cell_power()
                self.set_fuel_cell_mass()

            # if user-provided values are passed,
            # they override the default values
            if "capacity" in self.energy_storage:
                self.override_battery_capacity()

            diff = (self["driving mass"].sum().values - old_driving_mass) / self[
                "driving mass"
            ].sum()

        self.calculate_ttw_energy()
        self.set_ttw_efficiency()

        self.set_range()

        if self.target_range:
            self.override_range()

        self.set_share_recuperated_energy()
        self.set_battery_fuel_cell_replacements()
        self.adjust_cost()

        self.set_electric_utility_factor()
        self.set_electricity_consumption()
        self.set_costs()
        self.set_hot_emissions()
        self.set_particulates_emission()
        self.set_noise_emissions()
        self.set_vehicle_mass()
        self.create_PHEV()
        if self.drop_hybrids:
            self.drop_hybrid()

        self.remove_energy_consumption_from_unavailable_vehicles()

    def set_battery_chemistry(self):
        # override default values for batteries
        # if provided by the user
        if "electric" not in self.energy_storage:
            self.energy_storage.update(
                {
                    "electric": {
                        x: "NMC-622"
                        for x in product(
                            ["BEV", "PHEV-e", "HEV-d", "HEV-p"],
                            self.array.coords["size"].values,
                            self.array.year.values,
                        )
                    },
                }
            )
        if "origin" not in self.energy_storage:
            self.energy_storage.update({"origin": "CN"})

    def adjust_cost(self) -> None:
        """
        This method adjusts costs of energy storage over time, to correct for the overly optimistic linear
        interpolation between years.

        """

        n_iterations = self.array.shape[-1]
        n_year = len(self.array.year.values)

        # If uncertainty is not considered, the cost factor equals 1.
        # Otherwise, a variability of +/-30% is added.

        if n_iterations == 1:
            cost_factor = 1
        else:
            if "reference" in self.array.value.values.tolist():
                cost_factor = np.ones((n_iterations, 1))
            else:
                cost_factor = np.random.triangular(0.7, 1, 1.3, (n_iterations, 1))

        # Correction of hydrogen tank cost, per kg
        # Correction of fuel cell stack cost, per kW
        if "FCEV" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (1.078e58 * np.exp(-6.32e-2 * self.array.year.values) + 3.43e2)
                * cost_factor,
                (1, n_year, n_iterations),
            )

            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (3.15e66 * np.exp(-7.35e-2 * self.array.year.values) + 2.39e1)
                * cost_factor,
                (1, n_year, n_iterations),
            )

        # Correction of energy battery system cost, per kWh
        list_batt = [
            i
            for i in ["BEV", "PHEV-e", "PHEV-c-p", "PHEV-c-d"]
            if i in self.array.powertrain
        ]
        if len(list_batt) > 0:
            self.array.loc[
                dict(powertrain=list_batt, parameter="energy battery cost per kWh")
            ] = np.reshape(
                (2.75e86 * np.exp(-9.61e-2 * self.array.year.values) + 5.059e1)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of power battery system cost, per kW
        list_pwt = [
            i
            for i in [
                "ICEV-p",
                "ICEV-d",
                "ICEV-g",
                "PHEV-c-p",
                "PHEV-c-d",
                "FCEV",
                "HEV-p",
                "HEV-d",
            ]
            if i in self.array.powertrain
        ]

        if len(list_pwt) > 0:
            self.array.loc[
                dict(powertrain=list_pwt, parameter="power battery cost per kW")
            ] = np.reshape(
                (8.337e40 * np.exp(-4.49e-2 * self.array.year.values) + 11.17)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of combustion powertrain cost for ICEV-g
        if "ICEV-g" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="ICEV-g", parameter="combustion powertrain cost per kW")
            ] = np.clip(
                np.reshape(
                    (5.92e160 * np.exp(-0.1819 * self.array.year.values) + 26.76)
                    * cost_factor,
                    (1, n_year, n_iterations),
                ),
                None,
                100,
            )

    def calculate_ttw_energy(self) -> None:
        """
        This method calculates the energy required to operate auxiliary services as well
        as to move the car. The sum is stored under the parameter label "TtW energy" in :attr:`self.array`.

        """

        self.energy = self.ecm.motive_energy_per_km(
            driving_mass=self["driving mass"],
            rr_coef=self["rolling resistance coefficient"],
            drag_coef=self["aerodynamic drag coefficient"],
            frontal_area=self["frontal area"],
            electric_motor_power=self["electric power"],
            engine_power=self["power"],
            recuperation_efficiency=self["recuperation efficiency"],
            aux_power=self["auxiliary power demand"],
            battery_charge_eff=self["battery charge efficiency"],
            battery_discharge_eff=self["battery discharge efficiency"],
            fuel_cell_system_efficiency=self["fuel cell system efficiency"],
        )

        self.energy = self.energy.assign_coords(
            {
                "powertrain": self.array.powertrain,
                "year": self.array.year,
                "size": self.array.coords["size"],
                "value": self.array.coords["value"],
            }
        )

        if self.energy_consumption:
            self.override_ttw_energy()

        distance = self.energy.sel(parameter="velocity").sum(dim="second") / 1000

        self["transmission efficiency"] = (
            np.ma.array(
                self.energy.loc[dict(parameter="transmission efficiency")],
                mask=self.energy.loc[dict(parameter="power load")] == 0,
            )
            .mean(axis=0)
            .T
        )

        self["engine efficiency"] = (
            (
                (self.energy.sel(parameter="motive energy at wheels").sum(dim="second"))
                / (self.energy.sel(parameter="motive energy").sum(dim="second"))
            ).T
            / self["transmission efficiency"]
            / np.where(
                self["fuel cell system efficiency"] > 0,
                self["fuel cell system efficiency"],
                1,
            )
        )

        _o = lambda x: np.where((x == 0) | (x == np.nan), 1, x)

        if self.engine_efficiency is not None:
            print("Engine efficiency is being overridden.")
            for key, val in self.engine_efficiency.items():
                pwt, size, year = key
                if (
                    (val is not None)
                    & (pwt in self.array.powertrain.values)
                    & (year in self.array.year.values)
                    & (size in self.array["size"].values)
                ):
                    self.array.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="engine efficiency",
                        )
                    ] = float(val)

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="engine efficiency",
                        )
                    ] = float(val) * np.where(
                        self.energy.loc[
                            dict(
                                parameter="power load",
                                powertrain=pwt,
                                size=size,
                                year=year,
                            )
                        ]
                        == 0,
                        0,
                        1,
                    )

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="motive energy",
                        )
                    ] = self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="motive energy at wheels",
                        )
                    ] / (
                        _o(
                            self.energy.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="engine efficiency",
                                )
                            ]
                        )
                        * _o(
                            self.energy.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="transmission efficiency",
                                )
                            ]
                        )
                        * _o(
                            self.array.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="fuel cell system efficiency",
                                )
                            ]
                        )
                    )

        if self.transmission_efficiency is not None:
            print("Transmission efficiency is being overridden.")
            for key, val in self.transmission_efficiency.items():
                pwt, size, year = key

                if (
                    (val is not None)
                    & (pwt in self.array.powertrain.values)
                    & (year in self.array.year.values)
                    & (size in self.array["size"].values)
                ):
                    self.array.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="transmission efficiency",
                        )
                    ] = float(val)

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="transmission efficiency",
                        )
                    ] = float(val) * np.where(
                        self.energy.loc[
                            dict(
                                parameter="power load",
                                powertrain=pwt,
                                size=size,
                                year=year,
                            )
                        ]
                        == 0,
                        0,
                        1,
                    )

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="motive energy",
                        )
                    ] = self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="motive energy at wheels",
                        )
                    ] / (
                        _o(
                            self.energy.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="engine efficiency",
                                )
                            ]
                        )
                        * _o(
                            self.energy.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="transmission efficiency",
                                )
                            ]
                        )
                        * _o(
                            self.array.loc[
                                dict(
                                    powertrain=pwt,
                                    size=size,
                                    year=year,
                                    parameter="fuel cell system efficiency",
                                )
                            ]
                        )
                    )

        self["TtW energy"] = (
            self.energy.sel(
                parameter=["motive energy", "auxiliary energy", "recuperated energy"]
            ).sum(dim=["second", "parameter"])
            / distance
        ).T

        # saved_TtW_energy_by_recuperation = recuperated energy * electric motor efficiency * electric transmission efficency / (engine efficiency * transmission efficiency)

        self["TtW energy"] += (
            (
                self.energy.sel(parameter="recuperated energy").sum(dim="second")
                / distance
            ).T
            * self.array.sel(powertrain="BEV", parameter="engine efficiency")
            * self.array.sel(powertrain="BEV", parameter="transmission efficiency")
            / (
                self["engine efficiency"]
                * self["transmission efficiency"]
                * np.where(
                    self["fuel cell system efficiency"] == 0,
                    1,
                    self["fuel cell system efficiency"],
                )
            )
        )

        self["TtW energy, combustion mode"] = self["TtW energy"] * (
            self["combustion power share"] > 0
        )
        self["TtW energy, electric mode"] = self["TtW energy"] * (
            self["combustion power share"] == 0
        )

        self["auxiliary energy"] = (
            self.energy.sel(parameter="auxiliary energy").sum(dim="second") / distance
        ).T

    def set_vehicle_mass(self) -> None:
        """
        Define ``curb mass``, ``driving mass``, and ``total cargo mass``.

            * `curb mass <https://en.wikipedia.org/wiki/Curb_weight>`__ is the mass of the vehicle and fuel, without people or cargo.
            * ``total cargo mass`` is the mass of the cargo and passengers.
            * ``driving mass`` is the ``curb mass`` plus ``total cargo mass``.

        .. note::
            driving mass = total cargo mass + driving mass

        """

        self["curb mass"] = self["glider base mass"] * (1 - self["lightweighting"])

        curb_mass_includes = [
            "fuel mass",
            "charger mass",
            "converter mass",
            "inverter mass",
            "power distribution unit mass",
            # Updates with set_components_mass
            "combustion engine mass",
            # Updates with set_components_mass
            "electric engine mass",
            # Updates with set_components_mass
            "powertrain mass",
            "fuel cell stack mass",
            "fuel cell ancillary BoP mass",
            "fuel cell essential BoP mass",
            "battery cell mass",
            "battery BoP mass",
            "fuel tank mass",
        ]

        self["curb mass"] += self[curb_mass_includes].sum(axis=2)

        if self.target_mass:
            self.override_vehicle_mass()

        self["total cargo mass"] = (
            self["average passengers"] * self["average passenger mass"]
            + self["cargo mass"]
        )
        self["driving mass"] = self["curb mass"] + self["total cargo mass"]

    def set_electric_utility_factor(self) -> None:
        """Set the electric utility factor according to a sampled values in Germany (ICTT 2022)
        https://theicct.org/wp-content/uploads/2022/06/real-world-phev-use-jun22-1.pdf

        Real-world range in simulation 20 km 30 km 40 km 50 km 60 km 70 km 80 km
        Observed UF for Germany (Sample-size weighted regression ± 2 standard errors)
        Observed UF private (in %) 30±2 41±2 50±3 58±3 65±3 71±3 75±3

        which correlated the share of km driven in electric-mode to
        the capacity of the battery
        (the range that can be driven in battery-depleting mode).

        The argument `uf` is used to override this relation, if needed.
        `uf` must be a ratio between 0 and .75, for each."""

        if "PHEV-e" in self.array.coords["powertrain"].values:
            if self.electric_utility_factor is None:
                self.array.loc[
                    dict(powertrain="PHEV-e", parameter="electric utility factor")
                ] = np.clip(
                    np.interp(
                        self.array.loc[dict(powertrain="PHEV-e", parameter="range")],
                        [0, 20, 30, 40, 50, 60, 70, 80, 100, 120, 200],
                        [0, 0.13, 0.18, 0.23, 0.28, 0.30, 0.35, 0.40, 0.45, 0.5, 0.75],
                    ),
                    0,
                    0.75,
                )
            else:
                for key, val in self.electric_utility_factor.items():
                    if (
                        "PHEV-e" in self.array.powertrain.values
                        and key in self.array.year.values
                    ):
                        self.array.loc[
                            dict(
                                powertrain="PHEV-e",
                                parameter="electric utility factor",
                                year=key,
                            )
                        ] = val

    def set_costs(self) -> None:
        """
        Calculate the different cost types.
        :return:
        """
        self["glider cost"] = (
            self["glider base mass"] * self["glider cost slope"]
            + self["glider cost intercept"]
        )
        self["lightweighting cost"] = (
            self["glider base mass"]
            * self["lightweighting"]
            * self["glider lightweighting cost per kg"]
        )
        self["electric powertrain cost"] = (
            self["electric powertrain cost per kW"] * self["electric power"]
        )
        self["combustion powertrain cost"] = (
            self["combustion power"] * self["combustion powertrain cost per kW"]
        )
        self["fuel cell cost"] = self["fuel cell power"] * self["fuel cell cost per kW"]
        self["power battery cost"] = (
            self["battery power"] * self["power battery cost per kW"]
        )
        self["energy battery cost"] = (
            self["energy battery cost per kWh"] * self["electric energy stored"]
        )
        self["fuel tank cost"] = self["fuel tank cost per kg"] * self["fuel mass"]
        # Per km
        self["energy cost"] = self["energy cost per kWh"] * self["TtW energy"] / 3600

        # For battery, need to divide cost of electricity
        # at battery by efficiency of charging
        # to get costs at the "wall socket".

        _ = lambda x: np.where(x == 0, 1, x)
        self["energy cost"] /= _(self["battery charge efficiency"])

        self["component replacement cost"] = (
            self["energy battery cost"] * self["battery lifetime replacements"]
            + self["fuel cell cost"] * self["fuel cell lifetime replacements"]
        )

        with open(DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            to_markup = yaml.safe_load(stream)["markup"]

        self[to_markup] *= self["markup factor"]

        # calculate costs per km:
        self["lifetime"] = self["lifetime kilometers"] / self["kilometers per year"]

        with open(DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            purchase_cost_params = yaml.safe_load(stream)["purchase"]

        self["purchase cost"] = self[purchase_cost_params].sum(axis=2)
        # per km
        amortisation_factor = self["interest rate"] + (
            self["interest rate"]
            / (
                (np.array(1) + self["interest rate"]) ** self["lifetime kilometers"]
                - np.array(1)
            )
        )
        self["amortised purchase cost"] = (
            self["purchase cost"] * amortisation_factor / self["kilometers per year"]
        )

        # per km
        self["maintenance cost"] = (
            self["maintenance cost per glider cost"]
            * self["glider cost"]
            / self["kilometers per year"]
        )

        # simple assumption that component replacement
        # occurs at half of life.
        self["amortised component replacement cost"] = (
            (
                self["component replacement cost"]
                * (
                    (np.array(1) - self["interest rate"]) ** self["lifetime kilometers"]
                    / 2
                )
            )
            * amortisation_factor
            / self["kilometers per year"]
        )

        self["total cost per km"] = (
            self["energy cost"]
            + self["amortised purchase cost"]
            + self["maintenance cost"]
            + self["amortised component replacement cost"]
        )

    def remove_energy_consumption_from_unavailable_vehicles(self):
        """
        This method sets the energy consumption of vehicles that are not available to zero.
        """

        # we flag cars that have a range inferior to 100 km
        # and also BEVs, PHEVs and FCEVs from before 2013
        self["TtW energy"] = np.where((self["range"] < 100), 0, self["TtW energy"])

        pwts = [
            pt
            for pt in [
                "BEV",
                "PHEV-e",
                "PHEV-c-p",
                "PHEV-c-d",
                "FCEV",
                "PHEV-p",
                "PHEV-d",
                "HEV-d",
                "HEV-p",
            ]
            if pt in self.array.coords["powertrain"].values
        ]

        years = [y for y in self.array.year.values if y < 2013]

        if years:
            self.array.loc[
                dict(
                    parameter="TtW energy",
                    powertrain=pwts,
                    year=years,
                )
            ] = 0

        # and also Micro cars other than BEVs
        if "Micro" in self.array.coords["size"].values:
            self.array.loc[
                dict(
                    parameter="TtW energy",
                    powertrain=[
                        pt
                        for pt in [
                            "PHEV-e",
                            "PHEV-c-p",
                            "PHEV-c-d",
                            "FCEV",
                            "PHEV-p",
                            "PHEV-d",
                            "HEV-d",
                            "HEV-p",
                            "ICEV-p",
                            "ICEV-d",
                            "ICEV-g",
                        ]
                        if pt in self.array.coords["powertrain"].values
                    ],
                    size="Micro",
                )
            ] = 0

            # replace Nans with zeros
            self.array.loc[dict(size="Micro")] = self.array.loc[
                dict(size="Micro")
            ].fillna(0)

        if "BEV" in self.array.coords["powertrain"].values:
            # set the `TtW energy` of BEV vehicles before 2010 to zero
            self.array.loc[
                dict(
                    powertrain="BEV",
                    year=slice(None, 2010),
                    parameter="TtW energy",
                )
            ] = 0



import numpy as np
import pytest

from carculator import (
    CarInputParameters,
    CarModel,
    InventoryCar,
    fill_xarray_from_input_parameters,
)

# generate vehicle parameters
cip = CarInputParameters()
cip.static()

# fill in array with vehicle parameters
scope = {"powertrain": ["ICEV-d", "ICEV-p", "BEV"], "size": ["Medium"]}
_, array = fill_xarray_from_input_parameters(cip, scope=scope)

# build CarModel object
cm = CarModel(array, cycle="WLTC")
# build vehicles
cm.set_all()


def test_scope():
    """Test if scope works as expected"""

    # generate vehicle parameters
    cip = CarInputParameters()
    cip.static()

    # fill in array with vehicle parameters
    scope = {"powertrain": ["ICEV-d", "ICEV-p", "BEV"], "size": ["Medium"]}
    _, array = fill_xarray_from_input_parameters(cip, scope=scope)

    # build CarModel object
    cm = CarModel(array, cycle="WLTC")
    # build vehicles
    cm.set_all()

    ic = InventoryCar(
        cm,
        method="recipe",
        indicator="midpoint",
    )
    results = ic.calculate_impacts()

    assert "Large" not in results.coords["size"].values
    assert "FCEV" not in results.coords["powertrain"].values


def test_plausibility_of_GWP():
    """Test if GWP scores make sense"""

    for method in ["recipe", "ilcd"]:
        ic = InventoryCar(cm, method=method, indicator="midpoint")
        results = ic.calculate_impacts()

        if method == "recipe":
            m = "climate change"
        else:
            m = "climate change - climate change total"

        gwp_icev = results.sel(
            impact_category=m,
            powertrain=["ICEV-d", "ICEV-p"],
            value=0,
            year=2020,
            size="Medium",
        )

        # Are the medium ICEVs between 0.28 and 0.35 kg CO2-eq./vkm?

        if method == "recipe":
            assert (gwp_icev.sum(dim="impact") > 0.24).all() and (
                gwp_icev.sum(dim="impact") < 0.34
            ).all(), gwp_icev.sum(dim="impact")

            # Are the medium ICEVs direct emissions between 0.13 and  0.18 kg CO2-eq./vkm?
            assert (gwp_icev.sel(impact="direct - exhaust") > 0.13).all() and (
                gwp_icev.sel(impact="direct - exhaust") < 0.19
            ).all(), gwp_icev.sel(impact="direct - exhaust")

        # Is the GWP score for batteries of Medium BEVs between 0.025 and 0.035 kg Co2-eq./vkm?
        gwp_bev = results.sel(
            impact_category=m, powertrain="BEV", value=0, year=2020, size="Medium"
        )

        assert (gwp_bev.sel(impact="energy storage") > 0.02).all() and (
            gwp_bev.sel(impact="energy storage") < 0.04
        ).all()

        assert gwp_bev.sel(impact="direct - exhaust") == 0


def test_fuel_blend():
    """Test if fuel blends defined by the user are considered"""

    bc = {
        "petrol": {
            "primary": {
                "type": "petrol",
                "share": [0.9, 0.9, 0.9, 0.9, 0.9, 0.9],
            },
        },
        "diesel": {
            "primary": {
                "type": "diesel",
                "share": [0.93, 0.93, 0.93, 0.93, 0.93, 0.93],
            },
        },
        "hydrogen": {
            "primary": {
                "type": "hydrogen - electrolysis - PEM",
                "share": [0.9, 0.9, 0.9, 0.9, 0.9, 0.9],
            },
        },
        "methane": {
            "primary": {
                "type": "methane - biomethane - sewage sludge",
                "share": [
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
            }
        },
    }

    cm = CarModel(array, cycle="WLTC", fuel_blend=bc)
    cm.set_all()

    assert np.array_equal(
        cm.fuel_blend["petrol"]["primary"]["share"],
        np.array(
            [
                0.9,
                0.9,
                0.9,
                0.9,
                0.9,
                0.9,
            ]
        ),
    )

    assert np.array_equal(
        cm.fuel_blend["diesel"]["primary"]["share"],
        np.array(
            [
                0.93,
                0.93,
                0.93,
                0.93,
                0.93,
                0.93,
            ]
        ),
    )
    assert np.array_equal(
        cm.fuel_blend["methane"]["primary"]["share"], np.array([1, 1, 1, 1, 1, 1])
    )
    assert np.allclose(
        np.sum(cm.fuel_blend["methane"]["secondary"]["share"]), np.zeros(6)
    )

    for fuels in [
        ("petrol", "diesel", "hydrogen - electrolysis - PEM", "methane"),
        (
            "petrol - bioethanol - wheat straw",
            "diesel - biodiesel - palm oil",
            "hydrogen - smr - natural gas",
            "methane - biomethane - sewage sludge",
        ),
        (
            "petrol - bioethanol - forest residues",
            "diesel - biodiesel - rapeseed oil",
            "hydrogen - smr - natural gas with CCS",
            "methane - synthetic - coal",
        ),
        (
            "petrol - bioethanol - maize starch",
            "diesel - biodiesel - cooking oil",
            "hydrogen - wood gasification",
            "methane - synthetic - biological",
        ),
        (
            "petrol - synthetic - methanol - electrolysis - energy allocation",
            "diesel - synthetic - FT - coal - economic allocation",
            "hydrogen - atr - biogas",
            "methane - synthetic - biological - MSWI",
        ),
        (
            "petrol - synthetic - methanol - cement - energy allocation",
            "diesel - synthetic - methanol - cement - economic allocation",
            "hydrogen - wood gasification with CCS",
            "methane - synthetic - electrochemical - MSWI",
        ),
    ]:
        bc["petrol"]["primary"]["type"] = fuels[0]
        bc["diesel"]["primary"]["type"] = fuels[1]
        bc["hydrogen"]["primary"]["type"] = fuels[2]
        bc["methane"]["primary"]["type"] = fuels[3]

        cm = CarModel(array, cycle="WLTC", fuel_blend=bc)
        cm.set_all()
        ic = InventoryCar(cm)
        ic.calculate_impacts()


def test_countries():
    """Test that calculation works with all countries"""
    for c in [
        "AO",
        "AT",
        "AU",
        "BE",
    ]:
        ic = InventoryCar(
            cm,
            method="recipe",
            indicator="midpoint",
            background_configuration={
                "country": c,
                "energy storage": {"electric": {"type": "NMC-622"}, "origin": c},
            },
        )
        ic.calculate_impacts()


def test_endpoint():
    """Test if the correct impact categories are considered"""
    ic = InventoryCar(cm, method="recipe", indicator="endpoint")
    results = ic.calculate_impacts()
    assert "human health" in [i.lower() for i in results.impact_category.values]
    assert len(results.impact_category.values) == 4
    #
    #     """Test if it errors properly if an incorrect method type is give"""
    with pytest.raises(ValueError) as wrapped_error:
        ic = InventoryCar(cm, method="recipe", indicator="endpint")
        ic.calculate_impacts()
    assert wrapped_error.type == ValueError


def test_sulfur_concentration():
    ic = InventoryCar(cm, method="recipe", indicator="endpoint")
    ic.get_sulfur_content(location="FR", fuel="diesel")


def test_custom_electricity_mix():
    """Test if a wrong number of electricity mixes throws an error"""

    # Passing four mixes instead of 6
    mix_1 = np.zeros((5, 15))
    mix_1[:, 0] = 1

    mixes = [mix_1]

    for mix in mixes:
        with pytest.raises(ValueError) as wrapped_error:
            InventoryCar(
                cm,
                method="recipe",
                indicator="endpoint",
                background_configuration={"custom electricity mix": mix},
            )
        assert wrapped_error.type == ValueError


def test_export_to_bw():
    """Test that inventories export successfully"""
    ic = InventoryCar(cm, method="recipe", indicator="endpoint")
    #

    for b in ("3.9",):
        ic.export_lci(
            ecoinvent_version=b,
        )


def test_export_to_excel():
    """Test that inventories export successfully to Excel/CSV"""
    # generate vehicle parameters
    cip = CarInputParameters()
    cip.static()

    # fill in array with vehicle parameters
    scope = {"powertrain": ["ICEV-d", "ICEV-p", "BEV"], "size": ["Medium"]}
    _, array = fill_xarray_from_input_parameters(cip, scope=scope)

    # build CarModel object
    cm = CarModel(array, cycle="WLTC")
    # build vehicles
    cm.set_all()
    ic = InventoryCar(cm, method="recipe", indicator="endpoint")

    for s in ("brightway2", "simapro"):
        for d in ("file", "bw2io"):
            ic.export_lci(
                ecoinvent_version="3.9",
                format=d,
                software=s,
                directory="directory",
            )
